use std::path::PathBuf;
use std::sync::{self, Arc, Mutex};
use std::convert::TryInto;

use thiserror::Error;
use url::Host;
use rocket::{get, Route, routes, State};
use rocket::outcome::try_outcome;
use rocket::request::{FromRequest, Request, Outcome};

pub fn routes() -> Vec<Route> {
    routes![static_website]
}

#[derive(PartialEq, Clone, Debug, Error)]
pub enum Error {
    #[error("Can't lock an underling mutex")]
    Lock,
    #[error("Malformed subdomain, did you put '.' in it ?")]
    MalformedSubdomain,
    #[error("IP host are not supported")]
    IPHost,
}

type Result<T, E = Error> = core::result::Result<T, E>;

#[derive(Debug, Clone, PartialEq)]
pub(crate) struct Website {
    path: PathBuf,
    host: Host,
}

impl Website {
    /// Creates a website from a given subdomain and a base_host
    ///
    /// # Errors
    ///
    /// If your subdomain contains dots, an Error::MalformedUrl is returned
    /// Error::IPHost is returned if the given host is an IP
    fn from_subdomain<S: AsRef<str>>(subdomain: S,  base_host: &Host) -> Result<Self> {
        // TODO: implement the path creation
        let path = "".into();
        if subdomain.as_ref().contains(".") {
            return Err(Error::MalformedSubdomain);
        }
        let host = match base_host {
            Host::Domain(base_host) => format!("{}.{}", subdomain.as_ref(), base_host.to_string()),
            _ => return Err(Error::IPHost),
        };
        let host = Host::parse(&host).map_err(|_| Error::MalformedSubdomain)?;
        Ok(Self {
            path,
            host,
        })
    }
}

#[derive(Debug, Clone)]
pub(crate) struct Hosts {
    hosts: Arc<Mutex<Vec<Website>>>,
    base_host: Host,
}

impl Hosts {

    pub fn new(base_host: Host) -> Self {
        Self{
            hosts: Arc::new(Mutex::new(vec![])),
            base_host,
        }
    }

    pub fn contains_host(&self, host: &Host) -> Result<bool>{
        match self.hosts.lock() {
            Ok(hosts) => Ok(hosts.iter().any(|website| &website.host == host)),
            Err(_) => Err(Error::Lock),
        }
    }

    pub fn add_subdomain<S: AsRef<str>>(&self, subdomain: S) -> Result<()> {
        match &mut self.hosts.lock() {
            Ok(hosts) => {
                hosts.push(Website::from_subdomain(subdomain, &self.base_host)?);
                Ok(())
            },
            Err(_) => Err(Error::Lock),
        }
    }
}

pub struct Header<'a>(pub &'a str);

#[rocket::async_trait]
impl<'a> FromRequest<'a> for Header<'a> {
    type Error = ();

    async fn from_request(req: &'a Request<'_>) -> Outcome<Self, Self::Error> {
        let subdomain_hosts = try_outcome!(req.guard::<State<Hosts>>().await);
        match (req.headers().get_one("Host"), &subdomain_hosts) {
            // TODO: those unwrap should return Outcome Errors
            (Some(h), sh) if sh.contains_host(&Host::parse(h).unwrap()).unwrap() => Outcome::Success(Header(h)),
            (_, _) => return Outcome::Forward(()),
        }
    }
}

#[get("/<path..>")]
fn static_website(path: PathBuf, subdomain: Header) {
    println!("go go go to the subdomain {} and path {:?} !", subdomain.0, path);
}
